rm(list=ls())

randSeeds = sample.int(100, 5, replace = FALSE) ## used seed 19 to initially generate the seed numbers randomly

library(LogicReg)
library(ggplot2)
library(RVAideMemoire)
library(lubridate)
library(tibble)
library(reshape2)
library(stringr)
library(foreach)
library(vegan)
library(dplyr)
library(vctrs)

set.seed(2023) #Makes randomizations consistant between runs  

addBiomarker <- function(file_path) { ## adds biomarker
  # Initialize an empty list to store the extracted integers
  integer_list <- list()
  
  # Open the file for reading
  file_conn <- file(file_path, "r")
  
  # Read the lines from the file
  lines <- readLines(file_conn)
  
  # Close the file connection
  close(file_conn)
  
  # Open the file for writing (to overwrite the original file)
  file_conn <- file(file_path, "w")
  
  for (line in lines) {
    # Check if the line contains capital X followed by integers
    if (grepl("X\\d+", line)) {
      biomarkLine <- line
      # Extract integer values following capital X
      strLine <- str_extract_all(line, "X\\d+")
      strLine <- str_extract_all(strLine, "\\d+")
      bioIntList <- as.numeric(unlist(strLine))
      
      for (bioInt in bioIntList) {
        biomarkLine <- gsub(paste0("X", bioInt), colnames(bin)[bioInt], biomarkLine)
      }
      
      # Replace the original line with the modified biomarkLine
      cat(biomarkLine, file = file_conn, sep = "\n")
    } else {
      # If the line doesn't contain the pattern, write it as is to the file
      cat(line, file = file_conn, sep = "\n")
    }
  }
  
  # Close the file connection
  close(file_conn)
}

infile = "/Users/danyu/Desktop/yedS_ompC-rcsDB_emrKY_reduced_seqs.csv"
raw<-read.table(infile,sep=",",header=TRUE, stringsAsFactors = FALSE) ## makes everything tab-separated?

#Data cleaning steps
raw=as.data.frame(apply(raw, 2, function(x) as.factor(x)))
raw[raw=="TRUE"]<-"T"
raw$Sample=gsub(" ","",raw$Sample)
row.names(raw)=raw$Sample

#List of sample categories, for instance B=Beaver U=unknown.
#Also assumes that samples of respective category starts with this letter.
#ORIGINAL EXPANDED HOST LIST: "B", "C", "K", "D", "H", "M", "O", "P", "R", "S", "T", "W"
sources =c("B", "K", "H", "P")
animal = 'P' #Decides the host model to be trained              
host_partition=NULL

comparison=sources
labels=substr(raw[,1],1,1) ## so does this take the first letter of each isolate (the host) and create a vector indicating the host label
check=which(labels==animal) ## Returns the rows that matches the host letter

#creates a vector of a length equal to the number of samples
dataindex= vector(length=length(labels))

#This fills the dataindex with TRUE values if the samples belong to any category (H, R, B)
for (i in 1:length(labels)){
  print(i)
  if  (length(which(c(animal, comparison)==labels[i]))>0) {
    dataindex[i]=TRUE
  }
}

# eliminate rows not fitting with any host category (if any)
raw=raw[dataindex,]

#Creates a bool vector that tells if the sample belongs to modelhost  
outcome = labels == animal

#Section for identifying SNP nucleotides, by counting how many times each nucleotide occurs in a row
numbers=vector(length=dim(raw)[2]) ## this will create a vector of length equal to the number of columns (i.e., the length of the sequence to be analyzed)

#This checks in each column, how many times the values are same as the first row in that value.
#With the purpose to see if all rows are same in respective column.
for (i in 1:length(numbers)){
  numbers[i]= sum(raw[,i]== raw[1,i])  
}

#Convert the numeric list to boolean, depending if it is same length as total rows.
#This is used to find SNPs and then deselect columns with only one nucleotide (in next row).  
same=(numbers==dim(raw)[1])

## This selects SNP columns, and add a start column which indicate if modelhost or not
data=cbind(outcome,raw[!same])
data$outcome=as.factor(data$outcome)

### analysis procedure, construct holders for the data for logic regression
binA=NULL
binC=NULL
binG=NULL
binT=NULL
binM=NULL
first=TRUE

#Create boolean tables based on the presence abscense of nucleotides and gaps.
#binA will for instance go through
for (i in 1:length(colnames(data)) ){  #if fail, check if first column is sample name and move to second
  if (first) {
    binA=(data[i]=='A')
    binC=(data[i]=='C')
    binG=(data[i]=='G')
    binT=(data[i]=='T')
    binM=(data[i]=='-')
    first=FALSE
  }
  else  {
    binA <- cbind(binA, data[i]=='A')
    binC <- cbind(binC, data[i]=='C')
    binG <- cbind(binG, data[i]=='G')
    binT <- cbind(binT, data[i]=='T')
    binM <- cbind(binM, data[i]=='-')
  }
}

# assign names to the binary columns.
colnames(binA) <- paste(colnames(binA),'_A',sep='')
colnames(binC) <- paste(colnames(binC), '_C', sep='')
colnames(binG) <- paste(colnames(binG), '_G', sep='')                
colnames(binT) <- paste(colnames(binT), '_T', sep='')
colnames(binM) <- paste(colnames(binM), '_-', sep='')

# combine all the binary data by columns  
tmpbin=cbind(binA, binC, binG, binT, binM)

ix=0
bin=rep(NULL, dim(data)[1])

#Eliminate zeros columns, i.e. columns that never had a specificed nucleotid in any sample.  
#For instance column csgD_100 in binA, if csgD_100 never had an "A" in any of the samples.
for (i in 2:dim(tmpbin)[2]){
  if (sum(tmpbin[,i])>0 ){
    ix=ix+1
    bin=cbind(bin, tmpbin[,i])
    colnames(bin)[ix]=colnames(tmpbin)[i]
    print(paste("Stored column position:",  i))
  }  
}

"""
# logic regression parameters:
# type=3, using logistic regression model
# ntrees=2, number of logic trees to be fit
# select=1, type of model selection to be carried out: (1) fit a single model, (2) fit multiple models, (3) cross-validation, (4) null-model permutation test, (5) conditional permutation test, (6) a greedy stepwise algorithm, or (7) Monte Carlo Logic Regression (using MCMC).
# nleaves =10, maximum number of leaves to be fit in all trees combined

Detailed description about the logistic regression model and logreg library:
https://rdrr.io/cran/LogicReg/    #Library documentation
https://rdrr.io/cran/LogicReg/man/logreg.html #Description of the specific model function

#Documention by the authors describing  the theory (with examples) and implementation
https://research.fredhutch.org/content/dam/stripe/kooperberg/ingophd-logic.pdf  #
"""

#Fit a logistic regression model.  
SearchIterations <- 100000
#Decides the prediction strength threshold for when a sample is classified as a host.
hostBinaryCutoff = 0.5

#Settings for the algorithm searching for the local/global maximum on the likelihood curve
Annealing <- logreg.anneal.control(start=3.5, end=-4, iter=SearchIterations, update= 10000)
TreeControl <- logreg.tree.control(treesize=10, opers=1, minmass=0)    ## default setting

wbin = bin[grepl("^U.+",row.names(bin)),] #Store water samples bin
bin = bin[!(grepl("^U.+",row.names(bin))),] #Removes water sample prior to training data randomization

subID=sort(sample(1:dim(bin)[1],0.8*dim(bin)[1], replace=FALSE)) #Randomly select 80% of samples/rows for training set
allID=seq(1, dim(bin)[1],1) #Takes out all "row values"
oppID=allID[!(allID %in% subID)] #Takes out the remaining 20% of the samples for testing
sbin=bin[subID,] #Selects train data
tbin=bin[oppID,] #Select test data
soutcome=outcome[subID]

#Train/test the models
logregfit0<-logreg(soutcome, bin=sbin, type=3, select=1, ntrees=3, nleaves=15, anneal.control=Annealing, tree.control=TreeControl)
logregfit1<-logreg(soutcome, bin=sbin, type=3, select=2, ntrees=c(3), nleaves=c(20,30), oldfit = logregfit0,anneal.control=Annealing, tree.control=TreeControl)
logregfit2<-logreg(select = 3, oldfit = logregfit1)

plot(logregfit2) # Plot with test score from k-fold validations of all models

#Additional overview
#tapply(logregfit2$cvscores$test,list(logregfit2$cvscores$ntree,logregfit2$cvscores$nleaf),mean) #Mean of CV-score values per leave and tree
#tapply(logregfit2$cvscores$test,list(logregfit2$cvscores$ntree,logregfit2$cvscores$nleaf),sd) #SD of CV-score values per leave and tree
#logregfit1$alltrees
#summary(logregfit2)

#Choose the model based on the cv-test evaluation plot above
logregfit1$allscores  #Use this for the model selection in the line below
cvBest_model<-logregfit1$alltrees[[1]] # CHANGE Lowest cv-test score k-fold validation test.  
cvBestPredict=predict(logregfit1,newbin=tbin)

#Prediction / validation steps
bestModel=cvBestPredict$tr3.lf20 #CHANGE parameters to the used model
bestModel=cbind(as.data.frame(substr(labels[oppID], 1, 1)),bestModel)
names(bestModel) <-c("HostCatg","Prediction")
tapply(bestModel$Prediction, bestModel$HostCatg, mean)

bestModel$BinaryPred=ifelse(bestModel$Prediction>hostBinaryCutoff, 1,0)
bestModel$BinaryHost=ifelse(bestModel$HostCatg == animal, animal, "N")    

#Creates confusion matrix
conMx=as.data.frame(matrix(nrow = 2, ncol = 2), row.names = c(animal, "N"))
names(conMx) = c("0","1")

#Summarize the predictions of the model
truePos = length(bestModel$BinaryPred[which(bestModel$BinaryHost == animal & bestModel$BinaryPred==1)])
falseNeg = length(bestModel$BinaryPred[which(bestModel$BinaryHost == animal & bestModel$BinaryPred==0)])
trueNeg = length(bestModel$BinaryPred[which(bestModel$BinaryHost == "N" & bestModel$BinaryPred==0)])
falsePos = length(bestModel$BinaryPred[bestModel$BinaryHost == "N" & bestModel$BinaryPred==1])

conMx[1,1] = falseNeg
conMx[1,2] = truePos  
conMx[2,1] = trueNeg
conMx[2,2] = falsePos

#Accuracy
Accurarcy=(conMx[1,2]+conMx[2,1])/sum(conMx)

#Sensitivity TP/(TP/FN).
sensRes=conMx[1,2]/(conMx[1,2]+conMx[1,1])
#Specificy TN/(TN+FP)
specRes=conMx[2,1]/(conMx[2,1]+conMx[2,2])


#Change these! To the tree, leaf and host used in the model to be exported
nTree = "T3"
nLeaf = "L28a"
hostOrg = "P" ## CHANGE

#Name for the results summary file  
resSumry=paste0("ResultSumry_",hostOrg,nTree,nLeaf,"_",lubridate::as_date(Sys.time()),".txt")
#Determines where the export file goes
setwd("/Users/danyu/Desktop/")


#Creates a summary file by capturing print-messages and export them as a txt-file
sink(resSumry)
print(resSumry)
print(cvBest_model)
print(paste("Model accuracy:", Accurarcy))
print(paste("Model sensitivity:", sensRes))
print(paste("Model specificty:", specRes))
print("Confusion matrix prediction:")
print(conMx)
print("Mean logistic-regression prediction by host")
print(tapply(bestModel$Prediction, bestModel$HostCatg, mean))
perm.anova(Prediction ~ HostCatg, data=bestModel, nperm = 999,
           progress = TRUE)
addBiomarker(resSumry)
sink()

"""
###Water sample classification###
Similiary to the predict steps above, but using wbin (water samples)
Also does not validate the outcome, since 'true' true/false is unknown

Repeat this step for each host model trained (and change the output name!)
"""

#Determines the prediction strength needed to classify the sample.  
waterBinaryCutoff = 0.5

cvWestPredict=predict(logregfit1,newbin=wbin)

westModel=as.data.frame(cvWestPredict$tr3.lf15)  #Change to the used model
names(westModel) = "Prediction"
westModel$BinaryPred=ifelse(westModel$Prediction>waterBinaryCutoff, 1,0)
row.names(westModel)=row.names(wbin)

#Creates sample column at first position
westModel=add_column(westModel, row.names(westModel),.before = "Prediction")
waterSumry=paste0("WaterClassified_",resSumry)
names(westModel) = c("Sample", "Prediction", "BinaryPrediction")
write.table(westModel, waterSumry,fileEncoding = "UTF-8", sep = "\t", row.names = F)


"""
#Evaluates water samples based on the predicitons of previous models. - RUN AFTER MAKING ALL THE MODELS FIRST

The input of this section is summary of prediction values and binary classification
from previous models. This is put togther manually from the different model exports
in the previous section.
"""

setwd("/Users/danyu/Desktop/iterative_logreg_analysis/for_classifications_05-11-2023")
list.files()
df=read.csv("seed9939_classification_input.csv", fileEncoding = "UTF-8")

df[,2:4]=sapply(df[,2:4], function(x) as.numeric(gsub(",",".", x)))

bvCount=sum(df$BinaryPred_B)
hmCount=sum(df$BinaryPred_H)
rnCount=sum(df$BinaryPred_R)

allPredict=melt(df[,1:4], id.vars = c("Sample"), value.name = "Prediction")
allBinary=melt(df[,c(1,5:7)], id.vars = c("Sample"), value.name = "Binary")
allComb=cbind(allPredict, allBinary[,3])

allComb$variable=gsub("ModelPred_","",allComb$variable)
names(allComb)[4] = "BinaryPred"

#Loop that compares and classify the binary outcome of the other models.
hostCol = list()
newSampleCol = list()

#Add the specificity of respective model
for(smp in unique(allComb$Sample)){
  smpPos = grep(smp, allComb$Sample)
  tempDf = allComb[smpPos,]
  if(sum(tempDf$BinaryPred) == 0){
    hostCol = append(hostCol, "Undetermined")
    newSampleCol = append(unlist(newSampleCol), smp)
    print(paste(smp, "was classified as undetermined"))
    #Sys.sleep(0.5)
  }
}

for(smp in unique(allComb$Sample)){
  smpPos = grep(smp, allComb$Sample)
  tempDf = allComb[smpPos,]
  if(sum(tempDf$BinaryPred) == 1){
    tempDf = tempDf[grep(1, tempDf$BinaryPred),]
    hostCol = append(hostCol, unlist(tempDf$variable))
    newSampleCol = append(unlist(newSampleCol), smp)
    print(paste(smp, "was classified as", tempDf$variable))
    #Sys.sleep(0.5)
  }
}

"""
This loop compares water samples where mulitple models have given positive predictions.
It takes the prediction strength of the sample, and the models specifity, and combine
it into an evulation value.

This is compared to a 'difference value' which is set below (0.2 by default).

The highest evulation value of the sample is then substracted from all evulation values
(i.e all positive model predictions for that sample).

All evaulation values that remains above -0.2 after substraction, are the included
as a joint positive prediction. If only one value remains, the sample get the prediction
from that model.
"""
#Add the specificity of respective model
Host = c("B", "H", "R")
Specificity = c(0.9375, 0.9583, 0.9189)

hostFrame=data.frame(Host, Specificity)
hostFrame=hostFrame[order(hostFrame$Host),]

diffVal = 0.2

for(smp in unique(allComb$Sample)){
  smpPos = grep(smp, allComb$Sample)
  tempDf = allComb[smpPos,]
  tempDf = tempDf[order(tempDf$variable),]
  if(sum(tempDf$BinaryPred) > 1){
    print(paste("Comparing positive predictions in sample:", smp))
    tempDf=tempDf[grep(1, tempDf$BinaryPred),]
    posHosts=unique(stringr::str_sub(tempDf$variable, -1))
    regxPtn=paste0("(",unlist(posHosts),"$", ")",collapse = "|")
    tempDf$Evaluation=tempDf$Prediction[grep(regxPtn, tempDf$variable)]+hostFrame$Specificity[grep(regxPtn, hostFrame$Host)]
    print(tempDf)
    tempDf$Evaluation=tempDf$Evaluation-max(tempDf$Evaluation)
    tempDf=tempDf[tempDf$Evaluation>-diffVal,]
    prediction=paste0(tempDf$variable, collapse ="|")
    hostCol = append(hostCol,unlist(paste0(tempDf$variable, collapse ="|")))
    newSampleCol = append(unlist(newSampleCol), smp)
    print(paste(smp,"was labeled as", paste0(tempDf$variable, collapse ="|")))
    #Sys.sleep(3)
  }
}  

classFrame=data.frame(newSampleCol,unlist(hostCol))
names(classFrame)  = c("Sample","Classification")
classFrame=classFrame[order(classFrame$Sample),]

#Cleaning steps prior to export
classFrame$Classification = gsub("_R","ReinDeer", classFrame$Classification)
classFrame$Classification = gsub("_H","Human", classFrame$Classification)
classFrame$Classification = gsub("_B","Beaver", classFrame$Classification)
classFrame$Classification = gsub("Prediction","", classFrame$Classification)

#Checks that sample names/order is identical in original- and
#classification dataframe, then proceed to merge them and export.
if(sum(df$Sample %in% classFrame$Sample)==dim(df)[1]){
  exprtFrame=cbind(classFrame, df[,grep("^Prediction.+", names(df))])
  print(table(exprtFrame$Classification))
  write.csv(exprtFrame, "classified_WaterSamples.csv", row.names = FALSE)
}